﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Windows.Forms;
using Geotab.Checkmate.ObjectModel;
using Geotab.Web.Gms.ObjectModel.Api;
using MyAdminApiLib.Geotab.MyAdmin.MyAdminApi.ObjectModel;

namespace MyAdminApiExample
{
    public partial class MyAdminApiForm : Form
    {
        class AsyncArgument
        {
            public readonly WebServerInvoker Api;
            public readonly string MethodName;
            public readonly Dictionary<string, object> Parameters;
            public readonly Type ReturnType;

            public AsyncArgument(WebServerInvoker api, string methodName, Dictionary<string, object> parameters, Type returnType)
            {
                Api = api;
                MethodName = methodName;
                Parameters = parameters;
                ReturnType = returnType;
            }
        }

        class AsyncReturnResult
        {
            public readonly object ReturnResult;
            public readonly string ObjectName;

            public AsyncReturnResult(object returnResult, string objectName)
            {
                ReturnResult = returnResult;
                ObjectName = objectName;
            }
        }

        private readonly WebServerInvoker api;
        private const string ServerUrl = "https://myadminapi.geotab.com/v2/MyAdminApi.ashx";

        private ApiUser apiUser;
        private Guid apiKey;
        private Guid sessionId;
        private BackgroundWorker bw;

        public MyAdminApiForm()
        {
            InitializeComponent();
            api = new WebServerInvoker(null, 300000) {Url = ServerUrl};

            DateTime now = DateTime.Now;
            deviceContracts_ToDatePicker.Value = now;
            deviceContracts_FromDatePicker.Value = now;

            orderStatus_FromDateTimePicker.Value = now.AddMonths(-1);
            orderStatus_ToDateTimePicker.Value = now;

            activationHistory_StartDate.Value = now.AddMonths(-1);
            activationHistory_EndDate.Value = now;
        }

#region Authentication

        private void validateUserButton_Click(object sender, EventArgs e)
        {
            Cursor = Cursors.WaitCursor;
            try
            {
                treeView.Enabled = false;
                apiUser = null;
                apiKey = Guid.Empty;
                sessionId = Guid.Empty;
                userAdmin_AccountsListView.Clear();
                userAdmin_RolesListView.Clear();
                accountsList.Items.Clear();

                Authenticate();

                GetTerminationReasons();
                GetCountries();
                deviceBillings_GetAutoRequestsFromDate.Value = DateTime.Now.AddDays(-1);

                // Set Accounts
                if (apiUser.Accounts != null && apiUser.Accounts.Count > 0)
                {
                    foreach (ApiAccount account in apiUser.Accounts)
                    {
                        accountsList.Items.Add(account.AccountId);
                        ListViewItem item = userAdmin_AccountsListView.Items.Add(string.Format("{0} ({1})", account.AccountId, account.Currency.Code));
                        item.Tag = account;
                    }
                    accountsList.SelectedIndex = 0;
                }

                // Load selectable roles for user options
                if (apiUser.Roles != null && apiUser.Roles.Count > 0)
                {
                    foreach (ApiRole role in apiUser.Roles)
                    {
                        ListViewItem item = userAdmin_RolesListView.Items.Add(role.Name);
                        item.Tag = role;
                    }
                }
            }
            finally
            {
                Cursor = Cursors.Default;
            }
        }

        private bool Authenticate()
        {
            // Get Api Key
            var parameters = new Dictionary<string, object> { { "username", usernameBox.Text }, { "password", passwordBox.Text } };
            apiUser = api.Invoke("Authenticate", typeof(ApiUser), parameters) as ApiUser;
            if (apiUser == null)
            {
                return false;
            }
            splitContainer1.Enabled = true;

            // Set user ApiKey
            apiKey = apiUser.UserId;
            sessionId = apiUser.SessionId;
            userApiKeyBox.Text = apiKey.ToString();
            sessionIdBox.Text = sessionId.ToString();

            return true;
        }

#endregion

#region User Admin

        private void userAdmin_RegisterButton_Click(object sender, EventArgs e)
        {
            IList<string> accountNos = (from ListViewItem item in userAdmin_AccountsListView.Items where item.Checked select ((ApiAccount)item.Tag).AccountId).ToList();
            IList<string> roles = (from ListViewItem item in userAdmin_RolesListView.Items where item.Checked select ((ApiRole)item.Tag).Name).ToList();
            var parameters = new Dictionary<string, object>
            {
                { "apiKey", apiKey },
                { "sessionId", sessionId },
                { "username", userAdmin_EmailBox.Text },
                { "password", userAdmin_PasswordBox.Text },
                { "accountNos", accountNos },
                { "roles", roles }
            };
            var asyncArgument = new AsyncArgument(api, "RegisterNewUser", parameters, typeof(bool));
            BeginWorker(asyncArgument);
        }

        void UserAdmin_ValidateRegisterUser(object sender, EventArgs e)
        {
            userAdmin_RegisterButton.Enabled = (!string.IsNullOrEmpty(userAdmin_EmailBox.Text) && !string.IsNullOrEmpty(userAdmin_PasswordBox.Text) && userAdmin_AccountsListView.CheckedItems.Count > 0 && userAdmin_RolesListView.CheckedItems.Count > 0);
        }

#endregion

#region Device Admin
        private void deviceAdmin_LookupDeviceButton_Click(object sender, EventArgs e)
        {
            LoadBusyItemToTreeView(treeView);
            var parameters = new Dictionary<string, object> { { "apiKey", apiKey }, { "sessionId", sessionId }, { "serialNo", deviceAdmin_SerialNoBox.Text } };
            var asyncArgument = new AsyncArgument(api, "LookupDevice", parameters, typeof(ApiDeviceInstallResult));
            bw = new BackgroundWorker();
            bw.DoWork += bw_DoWork;
            bw.RunWorkerCompleted += bw_RunWorkerCompleted;
            bw.RunWorkerAsync(asyncArgument);
        }

        private void deviceAdmin_DevicePlansBtn_Click(object sender, EventArgs e)
        {
            var parameters = new Dictionary<string, object>
            {
                {"apiKey", apiKey},
                {"sessionId", sessionId}
            };
            AsyncArgument arg = new AsyncArgument(api, "GetDevicePlans", parameters, typeof(object));
            BeginWorker(arg);
        }

        private void deviceAdmin_WarrantyOptionsBtn_Click(object sender, EventArgs e)
        {
            var parameters = new Dictionary<string, object>
            {
                {"apiKey", apiKey},
                {"sessionId", sessionId}
            };
            AsyncArgument arg = new AsyncArgument(api, "GetWarrantyOptions", parameters, typeof(object));
            BeginWorker(arg);
        }

        private void deviceAdmin_SerialNoBox_TextChanged(object sender, EventArgs e)
        {
            deviceAdmin_LookupDeviceButton.Enabled = !string.IsNullOrEmpty(deviceAdmin_SerialNoBox.Text);
        }

#endregion

#region Products

        private void products_GetButton_Click(object sender, EventArgs e)
        {
            var forAccount = accountsList.SelectedItem.ToString();
            var parameters = new Dictionary<string, object> { { "apiKey", apiKey }, { "sessionId", sessionId }, { "forAccount", forAccount }, { "productCodeFilter", products_ProductCodeFilterBox.Text } };
            var asyncArgument = new AsyncArgument(api, "GetAvailableProducts", parameters, typeof(ApiProductAndPricing[]));
            BeginWorker(asyncArgument);
        }

#endregion

#region Orders

        private void order_AddProductButton_Click(object sender, EventArgs e)
        {
            int quantity;
            if (!int.TryParse(order_ItemQuantityBox.Text, out quantity))
            {
                return;
            }
            var productAndPricing = order_ProductCodeList.SelectedItem as ApiProductAndPricing;
            var code = order_ProductCodeList.Text.Trim().ToUpperInvariant();
            if (string.IsNullOrEmpty(code))
            {
                return;
            }
            var cost = 0.0;
            var monthlyCost = 0.0;
            if (productAndPricing != null)
            {
                cost = productAndPricing.Pricings[0].PurchasePrice;
                monthlyCost = productAndPricing.Pricings[0].MonthlyPrice;
            }
            var orderItem = new ApiOrderItem(code, quantity);
            if (order_ItemsListView.Columns.Count == 0)
            {
                order_ItemsListView.Columns.Add("Product Code", 120);
                order_ItemsListView.Columns.Add("Quantity", 80);
                order_ItemsListView.Columns.Add("Purchase Cost", 100);
                order_ItemsListView.Columns.Add("Monthly Cost", 100);
            }
            ListViewItem item = order_ItemsListView.Items.Add(orderItem.ProductCode);
            item.Tag = orderItem;
            item.SubItems.Add(orderItem.Quantity.ToString(CultureInfo.InvariantCulture));
            item.SubItems.Add((cost * quantity).ToString("c"));
            item.SubItems.Add((monthlyCost * quantity).ToString("c"));
        }

        private void order_ItemsListView_SelectedIndexChanged(object sender, EventArgs e)
        {
            order_ItemRemoveButton.Enabled = order_ItemsListView.SelectedItems.Count > 0;
        }

        private void order_ItemRemoveButton_Click(object sender, EventArgs e)
        {
            if (order_ItemsListView.SelectedItems.Count <= 0)
            {
                return;
            }
            foreach (ListViewItem item in order_ItemsListView.SelectedItems)
            {
                item.Remove();
            }
        }

        void LoadAccountSpecificLists()
        {
            if (accountsList.Items.Count == 0)
            {
                return;
            }
            Cursor = Cursors.WaitCursor;
            try
            {
                Dictionary<string, object> param;
                // Device plans
                PopulateDevicePlanList(order_ForAccountList.Text);
                
                // User Contacts
                List<ApiUserContact> userContacts = null;
                order_ShipToList.Items.Clear();
                deviceContracts_UserCompanyList.Items.Clear();
                updateContracts_UserContactList.Items.Clear();
                activationHistory_UserCompanyList.Items.Clear();
                Dictionary<int, ApiUserCompany> userCompanies = new Dictionary<int, ApiUserCompany>();
                param = new Dictionary<string, object>
                {
                    { "apiKey", apiKey },
                    { "sessionId", sessionId },
                    { "forAccount", order_ForAccountList.Text },
                    { "filter", "" },
                    { "activeOnly", true }
                };
                userContacts = (List<ApiUserContact>)api.Invoke("GetUserContacts", typeof(List<ApiUserContact>), param);

                if (userContacts == null)
                {
                    return;
                }
                foreach (ApiUserContact userContact in userContacts)
                {
                    order_ShipToList.Items.Add(userContact);
                    updateContracts_UserContactList.Items.Add(userContact);
                    if (!userCompanies.ContainsKey(userContact.UserCompany.Id))
                    {
                        userCompanies.Add(userContact.UserCompany.Id, userContact.UserCompany);
                    }
                }
                deviceContracts_UserCompanyList.Items.Add("");
                activationHistory_UserCompanyList.Items.Add("");
                foreach (ApiUserCompany userCompany in userCompanies.Values)
                {
                    deviceContracts_UserCompanyList.Items.Add(userCompany);
                    activationHistory_UserCompanyList.Items.Add(userCompany);
                }
                order_ShipToList.DisplayMember = "DisplayName";
                order_ShipToList.SelectedIndex = 0;
                updateContracts_UserContactList.DisplayMember = "DisplayName";
                updateContracts_UserContactList.SelectedIndex = 0;
                deviceContracts_UserCompanyList.DisplayMember = "Name";
                deviceContracts_UserCompanyList.SelectedIndex = 0;
                activationHistory_UserCompanyList.DisplayMember = "Name";
                activationHistory_UserCompanyList.SelectedIndex = 0;

                // Product Codes
                order_ProductCodeList.Items.Clear();

                param = new Dictionary<string, object>
                {
                    { "apiKey", apiKey },
                    { "sessionId", sessionId },
                    { "forAccount", order_ForAccountList.Text },
                    { "productCodeFilter", "" }
                };
                List<ApiProductAndPricing> products = (List<ApiProductAndPricing>)api.Invoke("GetAvailableProducts", typeof(List<ApiProductAndPricing>), param);
                if (products == null)
                {
                    return;
                }
                foreach (ApiProductAndPricing item in products)
                {
                    order_ProductCodeList.Items.Add(item);
                }
                order_ProductCodeList.DisplayMember = "Code";
                order_ProductCodeList.SelectedIndex = 0;

                // Shipping fees
                param = new Dictionary<string, object>
                {
                    { "apiKey", apiKey },
                    { "sessionId", sessionId }
                };
                List<ApiShippingFee> shippingFees = (List<ApiShippingFee>)api.Invoke("GetShippingFees", typeof(List<ApiShippingFee>), param);
                order_ShippingFeeList.Items.Clear();
                if (shippingFees == null)
                {
                    return;
                }
                foreach (ApiShippingFee shippingFee in shippingFees)
                {
                    order_ShippingFeeList.Items.Add(shippingFee);
                }
                order_ShippingFeeList.DisplayMember = "Name";
                order_ShippingFeeList.SelectedIndex = 0;

                List<ApiWarrantyOption> warrantyOptions = (List<ApiWarrantyOption>)api.Invoke("GetWarrantyOptions", typeof(List<ApiWarrantyOption>), param);
                order_WarrantyOptionList.Items.Clear();
                if (warrantyOptions == null)
                {
                    return;
                }
                foreach (ApiWarrantyOption warrantyOption in warrantyOptions)
                {
                    order_WarrantyOptionList.Items.Add(warrantyOption);
                }
                order_WarrantyOptionList.DisplayMember = "Name";
                order_WarrantyOptionList.SelectedIndex = 0;

                order_ItemRemoveButton.Enabled = false;
                order_ItemsListView.Clear();
            }
            finally
            {
                Cursor = Cursors.Default;
            }
        }

        private void PopulateDevicePlanList(string forAccount)
        {
            List<ApiDevicePlan> devicePlans = null;
            order_DevicePlanList.Items.Clear();
            deviceBillings_DevicePlanList.Items.Clear();
            deviceContracts_DevicePlanList.Items.Clear();
            deviceContracts_DevicePlanList.Items.Add("");
            var param = new Dictionary<string, object>
                {
                    { "apiKey", apiKey },
                    { "sessionId", sessionId },
                    { "forAccount", forAccount }
                };
            devicePlans = (List<ApiDevicePlan>)api.Invoke("GetDevicePlans", typeof(List<ApiDevicePlan>), param);
            if (devicePlans == null)
            {
                return;
            }
            foreach (ApiDevicePlan devicePlan in devicePlans)
            {
                if (devicePlan.ValidForOrder)
                {
                    order_DevicePlanList.Items.Add(devicePlan);
                }
                if (devicePlan.Level != 9999)
                {
                    deviceBillings_DevicePlanList.Items.Add(devicePlan);
                }
                deviceContracts_DevicePlanList.Items.Add(devicePlan);
            }
            order_DevicePlanList.DisplayMember = "Name";
            order_DevicePlanList.SelectedIndex = 0;
            deviceBillings_DevicePlanList.DisplayMember = "Name";
            deviceBillings_DevicePlanList.SelectedIndex = 0;
            deviceContracts_DevicePlanList.DisplayMember = "Name";
            deviceContracts_DevicePlanList.SelectedIndex = 0;
        }

        private void order_SubmitButton_Click(object sender, EventArgs e)
        {
            var devicePlan = (ApiDevicePlan)order_DevicePlanList.SelectedItem;
            var warrantyOption = (ApiWarrantyOption)order_WarrantyOptionList.SelectedItem;
            var userContact = order_ShipToList.SelectedItem as ApiUserContact;
            var shippingFee = (ApiShippingFee)order_ShippingFeeList.SelectedItem;
            var orderHeader = new ApiOrderHeader(order_ForAccountList.Text, order_PurchaseOrderNoBox.Text, order_ReferenceBox.Text, devicePlan.Level, warrantyOption.Id, userContact != null ? userContact.Id : -1, shippingFee.Id, order_PromoCodeBox.Text, string.Empty, order_CommentsBox.Text, null);
            IList<ApiOrderItem> orderItems = (from ListViewItem item in order_ItemsListView.Items select item.Tag).OfType<ApiOrderItem>().ToList();
            orderHeader.OrderItems = orderItems.ToArray();
            Cursor = Cursors.WaitCursor;
            try
            {
                api.Invoke("PostOrder", null, new Dictionary<string, object> { { "apiKey", apiKey }, { "sessionId", sessionId }, { "apiOrderHeader", orderHeader } });
                LoadResultsToTreeView(treeView, "Order successfully posted");
            }
            catch (SessionExpiredException ex)
            {
                // Session expired - re-authenticate and try again
                if (Authenticate())
                {
                    try
                    {
                        api.Invoke("PostOrder", null, new Dictionary<string, object> { { "apiKey", apiKey }, { "sessionId", sessionId }, { "apiOrderHeader", orderHeader } });
                        LoadResultsToTreeView(treeView, "Order successfully posted");
                    }
                    catch (Exception innerEx)
                    {
                        LoadResultsToTreeView(treeView, innerEx);
                    }

                }
            }
            catch (Exception ex)
            {
                LoadResultsToTreeView(treeView, ex);
            }
            finally
            {
                Cursor = Cursors.Default;
            }
        }

#endregion

#region Order Status

        private void orderStatus_SubmitButton_Click(object sender, EventArgs e)
        {
            var parameters = new Dictionary<string, object>
            {
                {"apiKey", apiKey},
                {"sessionId", sessionId},
                {"forAccount", orderStatus_ForAccountBox.Text},
                {"purchaseOrderNo", orderStatus_PoNumBox.Text},
                {"resellerReference", orderStatus_ResellerRefBox.Text},
                {"orderNo", orderStatus_OrderNumBox.Text},
                {"orderDateFrom", orderStatus_FromDateTimePicker.Value.Date.ToUniversalTime()},
                {"orderDateTo", orderStatus_ToDateTimePicker.Value.Date.AddDays(1).ToUniversalTime()},
                {"savedOrdersOnly", orderStatus_SavedOnlyCheckBox.Checked}
            };

            AsyncArgument arg = new AsyncArgument(api, "GetOnlineOrderStatus", parameters, typeof(object));
            BeginWorker(arg);
        }

#endregion

#region User Contacts

        private void userContacts_GetButton_Click(object sender, EventArgs e)
        {
            var forAccount = userContacts_ForAccountBox.Text;
            var parameters = new Dictionary<string, object> { { "apiKey", apiKey }, { "sessionId", sessionId }, { "forAccount", forAccount }, { "filter", userContacts_FilterBox.Text.Trim() }, { "activeOnly", userContacts_ActiveOnlyCheckBox.Checked } };
            var asyncArgument = new AsyncArgument(api, "GetUserContacts", parameters, typeof(ApiUserContact[]));
            BeginWorker(asyncArgument);
        }

        private void userContacts_SubmitBtn_Click(object sender, EventArgs e)
        {
            ApiUserContact userContact = new ApiUserContact();
            if (!string.IsNullOrEmpty(userContact_IdBox.Text))
            {
                userContact.Id = Convert.ToInt32(userContact_IdBox.Text);
            }
            else
            {
                userContact.Id = 0;
            }
            userContact.ContactName = userContacts_NameBox.Text;
            userContact.ContactEmail = userContacts_EmailBox.Text;
            userContact.Telephone1 = userContacts_Tel1Box.Text;
            userContact.Telephone2 = userContacts_Tel2Box.Text;
            userContact.Street1 = userContacts_Street1Box.Text;
            userContact.Street2 = userContacts_Street2Box.Text;
            userContact.City = userContacts_CityBox.Text;
            userContact.State = userContacts_StateList.Text;
            userContact.Country = userContacts_CountryList.SelectedItem.ToString();
            userContact.ZipCode = userContacts_ZipBox.Text;
            userContact.Active = userContacts_ActiveCheckBox.Checked;

            userContact.UserCompany = new ApiUserCompany();
            if (!string.IsNullOrEmpty(userContacts_CompanyIdBox.Text))
            {
                userContact.UserCompany.Id = Convert.ToInt32(userContacts_CompanyIdBox.Text);
            }
            else
            {
                userContact.UserCompany.Id = 0;
            }

            userContact.UserCompany.Name = userContacts_CompanyNameBox.Text;
            userContact.UserCompany.Active = userContacts_CompanyActiveCheckBox.Checked;
            userContact.UserCompany.Account = new ApiAccount() { AccountId = userContacts_ForAccountBox.Text };

            var parameters = new Dictionary<string, object>
            {
                {"apiKey", apiKey},
                {"sessionId", sessionId},
                {"userContact", userContact}
            };

            try
            {
                api.Invoke("EditUserContact", typeof(ApiUserContact), parameters);
            }
            catch (Exception ex)
            {

                MessageBox.Show(ex.Message);
            }

        }

        private void GetCountries()
        {
            var parameters = new Dictionary<string, object>
            {
                {"apiKey", apiKey},
                {"sessionId", sessionId}
            };

            List<string> countries = (List<string>)api.Invoke("GetCountries", typeof(List<string>), parameters);
            if (countries == null)
            {
                return;
            }
            userContacts_CountryList.Items.Clear();
            foreach (string country in countries)
            {
                userContacts_CountryList.Items.Add(country);
            }
            userContacts_CountryList.SelectedIndex = 0;
        }

        private void userContacts_CountryList_SelectedIndexChanged(object sender, EventArgs e)
        {
            string country = userContacts_CountryList.SelectedItem.ToString();
            var parameters = new Dictionary<string, object>
            {
                {"apiKey", apiKey},
                {"sessionId", sessionId},
                {"countryName", country}
            };

            List<string> states = (List<string>)api.Invoke("GetStates", typeof(List<string>), parameters);
            if (states == null)
            {
                return;
            }
            userContacts_StateList.Items.Clear();
            foreach (string state in states)
            {
                userContacts_StateList.Items.Add(state);
            }
            userContacts_StateList.SelectedIndex = 0;
        }

        #endregion

        #region Device Billings

        private void deviceBillings_ForAccountBox_Leave(object sender, EventArgs e)
        {
            order_ForAccountList.Text = deviceBillings_ForAccountBox.Text;
            PopulateDevicePlanList(deviceBillings_ForAccountBox.Text);
        }

        private void deviceBillings_ChangePlanBtn_Click(object sender, EventArgs e)
        {
            ApiDevicePlan devicePlan = (ApiDevicePlan)deviceBillings_DevicePlanList.SelectedItem;

            var parameters = new Dictionary<string, object>
            {
                {"apiKey", apiKey},
                {"sessionId", sessionId},
                {"serialNo", deviceBillings_SerialNoBox.Text },
                {"promoCode", deviceBillings_PromoCodeBox.Text },
                {"devicePlanId", devicePlan.Id },
                {"isImei", deviceBillings_ImeiCheckBox.Checked }
            };

            api.Invoke("ChangeDeviceBillingPlan", parameters);
        }

        private void GetTerminationReasons()
        {
            var parameters = new Dictionary<string, object>
            {
                {"apiKey", apiKey},
                {"sessionId", sessionId}
            };

            deviceBillings_TerminationReasonList.Items.Clear();
            List<ApiDeviceContractRequestReason> terminationReasons = (List<ApiDeviceContractRequestReason>)api.Invoke("GetTerminationReasons", typeof(List<ApiDeviceContractRequestReason>), parameters);

            if (terminationReasons == null)
            {
                return;
            }
            foreach (ApiDeviceContractRequestReason reason in terminationReasons)
            {
                if(reason.ParentId == null)
                {
                    deviceBillings_TerminationReasonList.Items.Add(reason);
                }
            }
            deviceBillings_TerminationReasonList.DisplayMember = "Name";
            deviceBillings_TerminationReasonList.SelectedIndex = 0;
        }

        private void UpdateSecondaryTerminationReasons(object sender, EventArgs e)
        {
            UpdateSecondaryTerminationReasons(((ApiDeviceContractRequestReason)deviceBillings_TerminationReasonList.SelectedItem).Id);
        }

        private void UpdateSecondaryTerminationReasons(int parentId)
        {
            var parameters = new Dictionary<string, object>
            {
                {"apiKey", apiKey},
                {"sessionId", sessionId},
                {"primaryRequestReasonId", parentId}
            };
            deviceBillings_SecondaryTerminationReasonList.Items.Clear();
            List<ApiDeviceContractRequestReason> terminationReasons = (List<ApiDeviceContractRequestReason>)api.Invoke("GetSecondaryTerminationReasons", typeof(List<ApiDeviceContractRequestReason>), parameters);
            if (terminationReasons == null)
            {
                return;
            }
            foreach (ApiDeviceContractRequestReason reason in terminationReasons)
            {
                deviceBillings_SecondaryTerminationReasonList.Items.Add(reason);
            }

            deviceBillings_SecondaryTerminationReasonList.DisplayMember = "Name";
            deviceBillings_SecondaryTerminationReasonList.SelectedIndex = 0;
        }

        private void deviceBillings_TerminateDeviceBtn_Click(object sender, EventArgs e)
        {
            ApiDeviceContractRequestReason reason = null;
            if (deviceBillings_SecondaryTerminationReasonList.Items.Count < 1)
            {
                reason = (ApiDeviceContractRequestReason)deviceBillings_TerminationReasonList.SelectedItem;
            }
            else
            {
                reason = (ApiDeviceContractRequestReason)deviceBillings_SecondaryTerminationReasonList.SelectedItem;
            } 
            
            var parameters = new Dictionary<string, object>
            {
                {"apiKey", apiKey},
                {"sessionId", sessionId},
                {"serialNo", deviceBillings_SerialNoBox.Text},
                {"reasonId", reason.Id},
                {"comments", deviceBillings_TerminationComments.Text },
                {"isImei", deviceBillings_ImeiCheckBox.Checked }
            };
            api.Invoke("TerminateDeviceBilling", parameters);
        }

        private void deviceBillings_GetAutoRequestsButton_Click(object sender, EventArgs e)
        {
            var parameters = new Dictionary<string, object>
            {
                {"apiKey", apiKey},
                {"sessionId", sessionId},
                {"forAccount", deviceDatabases_ForAccountBox.Text},
                {"serialNos", null },
                {"userCompanyIdFilter", null },
                {"fromDate", deviceBillings_GetAutoRequestsFromDate.Value },
                {"toDate", deviceBillings_GetAutoRequestsToDate.Value }
            };
            AsyncArgument arg = new AsyncArgument(api, "GetDeviceContractAutoRequests", parameters, typeof(object));
            BeginWorker(arg);
        }

        private void deviceBillings_GetTransactionsButton_Click(object sender, EventArgs e)
        {
            var forAccount = deviceBillings_ForAccountBox.Text;
            var period = deviceBillings_BillingMonthPicker.Value.Date;
            var parameters = new Dictionary<string, object> { { "apiKey", apiKey }, { "sessionId", sessionId }, { "forAccount", forAccount }, { "serialNoFilter", deviceBillings_SerialNoBox.Text }, { "monthFilter", period.Month }, { "yearFilter", period.Year } };
            try
            {
                int nextId = Convert.ToInt32(deviceBillings_NextIdBox.Text);
                parameters.Add("nextId", nextId);
            }
            catch
            {
            }
            var asyncArgument = new AsyncArgument(api, "GetDeviceContractTransactions", parameters, typeof(ApiDeviceContractTransaction[]));
            BeginWorker(asyncArgument);
        }

#endregion

#region Help Desk
        private void helpDesk_SubmitTicketBtn_Click(object sender, EventArgs e)
        {
            string subject = helpDesk_SubjectBox.Text;
            string description = helpDesk_DescriptionBox.Text;
            string[] fileToken = null;

            try
            {
                if (!string.IsNullOrEmpty(helpDesk_AttachmentBox.Text))
                {
                    FileStream file = File.Open(helpDesk_AttachmentBox.Text, FileMode.Open, FileAccess.Read);
                    byte[] attachment = new byte[file.Length];
                    file.Read(attachment, 0, (int) file.Length);
                    string contentType = "image/jpeg";
                    string fileName = helpDesk_AttachmentBox.Text.Substring(helpDesk_AttachmentBox.Text.LastIndexOf('\\') + 1);

                    var uploadParams = new Dictionary<string, object>
                    {
                        {"apiKey", apiKey},
                        {"sessionId", sessionId},
                        {"userEmail", helpDesk_EmailBox.Text},
                        {"file", attachment},
                        {"contentType", contentType},
                        {"fileName", fileName}
                    };

                    fileToken = new []{ api.Invoke("UploadSupportTicketFile", typeof(string), uploadParams) as string };
                }

                Dictionary<string, object> parameters;
                var serialNos = helpDesk_SerialNoBox.Text.Replace("\r\n", "\n").Split(new char[] { '\n' });
                var problemDescriptions = helpDesk_DeviceDescriptionBox.Text.Replace("\r\n", "\n").Split(new char[] { '\n' });

                if (!string.IsNullOrEmpty(helpDesk_TicketIdBox.Text))
                {
                    double ticketToken = Convert.ToDouble(helpDesk_TicketIdBox.Text);
                    parameters = new Dictionary<string, object>
                    {
                        {"apiKey", apiKey},
                        {"sessionId", sessionId},
                        {"userEmail", helpDesk_EmailBox.Text},
                        {"ticketToken", ticketToken},
                        {"body", description},
                        {"fileToken", fileToken},
                        {"resolveTicket", helpDesk_ResolveTicketCheckBox.Checked}
                    };
                    var ticketAudits =
                        api.Invoke("EditSupportTicket", typeof (IList<ApiTicketAudit>), parameters) as
                            IList<ApiTicketAudit>;
                }
                else
                {
                    parameters = new Dictionary<string, object>
                    {
                        {"apiKey", apiKey},
                        {"sessionId", sessionId},
                        {"userEmail", helpDesk_EmailBox.Text},
                        {"subject", subject},
                        {"description", description},
                        {"fileToken", fileToken},
                        {"serialNos", serialNos},
                        {"problemDescriptions", problemDescriptions}
                    };
                    ApiTicket ticket = new ApiTicket() {Subject = subject, Description = description};
                    ApiTicket returnTicket =
                        api.Invoke("CreateSupportTicket", typeof (ApiTicket), parameters) as ApiTicket;
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        private void helpDesk_BrowseButton_Click(object sender, EventArgs e)
        {
            OpenFileDialog fileDialog = new OpenFileDialog();
            fileDialog.Multiselect = false;
            if (fileDialog.ShowDialog() == DialogResult.OK)
            {
                string fileName = fileDialog.FileName;
                helpDesk_AttachmentBox.Text = fileName;
            }
        }

        private void helpDesk_GetTicketsBtn_Click(object sender, EventArgs e)
        {
            DateTime fromDate = DateTime.Now.AddDays(-1);
            DateTime toDate = DateTime.Now.AddDays(0);
            var parameters = new Dictionary<string, object> { { "apiKey", apiKey }, { "sessionId", sessionId }, { "userEmail", helpDesk_EmailBox.Text }, { "searchString", helpDesk_SearchTextBox.Text }, { "includeOrganization", helpDesk_IncludeOrganizationCheckBox.Checked }, { "includeOpen", helpDesk_IncludeOpenCheckBox.Checked }, { "includeClosed", helpDesk_IncludeClosedCheckBox.Checked }, { "fromDate", fromDate }, { "toDate", toDate } };

            AsyncArgument arg = new AsyncArgument(api, "GetSupportTickets", parameters, typeof(object));
            BeginWorker(arg);
        }

        private void helpDesk_DetailsBtn_Click(object sender, EventArgs e)
        {
            double ticketToken = Convert.ToDouble(helpDesk_TicketIdBox.Text);
            var parameters = new Dictionary<string, object>
            {
                {"apiKey", apiKey},
                {"sessionId", sessionId},
                {"userEmail", helpDesk_EmailBox.Text},
                {"ticketToken", ticketToken}
            };

            AsyncArgument arg = new AsyncArgument(api, "GetSupportTicketDetails", parameters, typeof(object));
            BeginWorker(arg);
        }

        private void helpDesk_Details_Validate(object sender, EventArgs e)
        {
            double d;
            if(double.TryParse(helpDesk_TicketIdBox.Text, out d))
            {
                helpDesk_DetailsBtn.Enabled = true;
            }
            else
            {
                helpDesk_DetailsBtn.Enabled = false;
            }
        }

#endregion
        
#region RMAs
        private void rma_GetRequestsBtn_Click(object sender, EventArgs e)
        {
            var parameters = new Dictionary<string, object>
            {
                {"apiKey", apiKey},
                {"sessionId", sessionId},
                {"forAccount", rma_ForAccountBox.Text}
            };

            AsyncArgument arg = new AsyncArgument(api, "GetRmaRequests", parameters, typeof(object));
            BeginWorker(arg);
        }

        private void rma_GetRequestItemsBtn_Click(object sender, EventArgs e)
        {
            if (!string.IsNullOrEmpty(rma_RequestIdBox.Text))
            {
                int id = Convert.ToInt32(rma_RequestIdBox.Text);
                ApiRmaRequest request = new ApiRmaRequest { Id = id };
                request.Account = new ApiAccount() { AccountId = rma_ForAccountBox.Text };

                var parameters = new Dictionary<string, object>
                {
                    {"apiKey", apiKey},
                    {"sessionId", sessionId},
                    {"forAccount", rma_ForAccountBox.Text},
                    {"rmaRequestId", id}
                };

                AsyncArgument arg = new AsyncArgument(api, "GetRmaRequestItems", parameters, typeof(object));
                BeginWorker(arg);
            }
        }

        private void rma_GetRmasBtn_Click(object sender, EventArgs e)
        {
            var parameters = new Dictionary<string, object>
            {
                {"apiKey", apiKey},
                {"sessionId", sessionId},
                {"forAccount", rma_ForAccountBox.Text}
            };
            if (!string.IsNullOrEmpty(rma_SerialNoBox.Text))
            {
                parameters.Add("deviceFilter", rma_SerialNoBox.Text);
            }
            if (!string.IsNullOrEmpty(rma_ResellerRefBox.Text))
            {
                parameters.Add("resellerReferenceFilter", rma_ResellerRefBox.Text);
            }
            AsyncArgument arg = new AsyncArgument(api, "GetRmas", parameters, typeof(object));
            BeginWorker(arg);
        }

        private void rma_SubmitRmaReqBtn_Click(object sender, EventArgs e)
        {
            var parameters = new Dictionary<string, object>
            {
                {"apiKey", apiKey},
                {"sessionId", sessionId},
                {"forAccount", rma_ForAccountBox.Text},
                {"activeOnly", true}
            };

            var userContacts = api.Invoke("GetUserContacts", typeof (ApiUserContact[]), parameters) as ApiUserContact[];

            if (userContacts != null && userContacts.Length > 0)
            {
                try
                {

                    var devices = rma_DeviceBox.Text.Replace("\r\n", "\n").Split(new char[] {'\n'});
                    var reasons = rma_ReasonBox.Text.Replace("\r\n", "\n").Split(new char[] {'\n'});
                    parameters = new Dictionary<string, object>
                    {
                        {"apiKey", apiKey},
                        {"sessionId", sessionId},
                        {"forAccount", rma_ForAccountBox.Text},
                        {"shipToId", userContacts[0].Id},
                        {"caseNo", rma_CaseNoBox.Text},
                        {"comments", rma_CreateRmaCommentsBox.Text},
                        {"resellerReference", rma_CreateRmaRefBox.Text},
                        {"itemsForRepair", devices},
                        {"reasonsForRepair", reasons}
                    };

                    var request = api.Invoke("CreateRmaRequest", typeof (ApiRmaRequest), parameters) as ApiRmaRequest;
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message);
                }
            }
        }

#endregion

#region Install Logs

        private void installLogs_LogInstallBtn_Click(object sender, EventArgs e)
        {
            double odometer = 0;
            Double.TryParse(installLogs_OdometerBox.Text, out odometer);

            var parameters = new Dictionary<string, object>
            {
                {"apiKey", apiKey},
                {"sessionId", sessionId},
                {"installerCompany", installLogs_CompanyNameBox.Text},
                {"installerName", installLogs_InstallerNameBox.Text},
                {"serialNo", installLogs_SerialNoBox.Text},
                {"assetNumber", installLogs_AssetNoBox.Text},
                {"odometer", odometer}
            };
            var result = api.Invoke("LogInstall", typeof(ApiDeviceInstallResult), parameters) as ApiDeviceInstallResult;
        }

        private void installLogs_LookupInstallLogBtn_Click(object sender, EventArgs e)
        {
            var parameters = new Dictionary<string, object>
            {
                {"apiKey", apiKey},
                {"sessionId", sessionId},
                {"installerFilter", installLogs_InstallerNameBox.Text},
                {"companyNameFilter", installLogs_CompanyNameBox.Text},
                {"serialNoFilter", installLogs_SerialNoBox.Text},
                {"assetNumber", installLogs_AssetNoBox.Text},
                {"fromDate", installLogs_InstallFromDate.Value.Date.ToUniversalTime()},
                {"toDate", installLogs_InstallToDate.Value.Date.ToUniversalTime().AddDays(1)}
            };

            AsyncArgument arg = new AsyncArgument(api, "GetInstallLogs", parameters, typeof(object));
            BeginWorker(arg);
        }

#endregion

#region Device Contracts

        private void deviceContracts_GetBtn_Click(object sender, EventArgs e)
        {
            var devices = deviceContracts_SerialNos.Text.Replace("\r\n", "\n").Split(new char[] { '\n' });
            var imeis = deviceContracts_Imeis.Text.Replace("\r\n", "\n").Split(new char[] { '\n' });

            var parameters = new Dictionary<string, object>
            {
                {"apiKey", apiKey},
                {"sessionId", sessionId},
                {"forAccount", deviceContracts_ForAccountBox.Text},
                {"fromDate", deviceContracts_FromDatePicker.Value.ToUniversalTime()},
                {"toDate", deviceContracts_ToDatePicker.Value.ToUniversalTime()},
                {"serialNos", devices},
                {"imeis", imeis },
                {"commentsFilter", deviceContracts_CommentsFilterBox.Text},
                {"purchaseOrderFilter", deviceContracts_PoFilterBox.Text},
                {"groupIdFilter", deviceContracts_GroupIdTextBox.Text},
                {"includeConnectInfo", deviceContracts_ConnectInfoCheck.Checked},
                {"showUnactivatedOnly", deviceContracts_UnactivatedCheckBox.Checked},
                {"showUnregisteredOnly", deviceContracts_UnregisteredCheckBox.Checked}
            };
            if (!string.IsNullOrEmpty(deviceContracts_NextIdBox.Text))
            {
                parameters.Add("nextId", Convert.ToInt32(deviceContracts_NextIdBox.Text));
            }

            ApiUserCompany userCompany = deviceContracts_UserCompanyList.SelectedItem as ApiUserCompany;
            if (userCompany != null)
            {
                parameters.Add("userCompanyId", userCompany.Id);
            }

            ApiDevicePlan devicePlan = deviceContracts_DevicePlanList.SelectedItem as ApiDevicePlan;
            if (devicePlan != null)
            {
                parameters.Add("devicePlanId", devicePlan.Id);
            }

            AsyncArgument arg = new AsyncArgument(api, "GetDeviceContracts", parameters, typeof(object));
            BeginWorker(arg);
        }

        private void updateContracts_UpdateCommentsCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            updateContracts_Comments.Enabled = updateContracts_UpdateCommentsCheckBox.Checked;
        }

        private void updateContracts_UpdateUserContactCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            updateContracts_UserContactList.Enabled = updateContracts_UpdateUserContactCheckBox.Checked;
        }

        private void updateContracts_UpdateActiveTracking_CheckedChanged(object sender, EventArgs e)
        {
            updateContracts_ActiveTrackingList.Enabled = updateContracts_UpdateActiveTracking.Checked;
        }

        private void updateContracts_UpdateGroupIdCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            updateContracts_GroupId.Enabled = updateContracts_UpdateGroupIdCheckBox.Checked;
        }

        private void updateContracts_SubmitButton_Click(object sender, EventArgs e)
        {
            ApiUserContact userContact = updateContracts_UserContactList.SelectedItem as ApiUserContact;
            int? userContactId = userContact == null ? (int?)null : userContact.Id;
            bool? activeTrackingDisallowed = null;
            if (updateContracts_ActiveTrackingList.SelectedIndex == 1)
            {
                activeTrackingDisallowed = false;
            }
            else if (updateContracts_ActiveTrackingList.SelectedIndex == 2)
            {
                activeTrackingDisallowed = true;
            }
            string[] serialNos = updateContracts_SerialNos.Lines;
            var parameters = new Dictionary<string, object>
            {
                {"apiKey", apiKey},
                {"sessionId", sessionId},
                {"serialNos", serialNos},
                {"updateComments", updateContracts_UpdateCommentsCheckBox.Checked},
                {"comments", updateContracts_Comments.Text},
                {"updateUserContact", updateContracts_UpdateUserContactCheckBox.Checked},
                {"updateGroupId", updateContracts_UpdateGroupIdCheckBox.Checked},
                {"groupId", updateContracts_GroupId.Text},
                {"userContactId", userContactId},
                {"updateActiveTracking", updateContracts_UpdateActiveTracking.Checked},
                {"activeTrackingDisallowed", activeTrackingDisallowed}
            };

            AsyncArgument arg = new AsyncArgument(api, "UpdateDeviceContracts", parameters, typeof(object));
            BeginWorker(arg);
        }

#endregion
        
#region Activation History
        private void activationHistory_SearchBtn_Click(object sender, EventArgs e)
        {
            var devices = activationHistory_SerialNosBox.Text.Replace("\r\n", "\n").Split(new char[] { '\n' });

            int? userCompanyId = null;
            ApiUserCompany userCompany = activationHistory_UserCompanyList.SelectedItem as ApiUserCompany;
            if (userCompany != null)
            {
                userCompanyId = userCompany.Id;
            }

            var parameters = new Dictionary<string, object>
            {
                {"apiKey", apiKey},
                {"sessionId", sessionId},
                {"forAccount", activationHistory_ForAccountBox.Text},
                {"serialNos", devices},
                {"userCompanyIdFilter", userCompanyId},
                {"fromDate", activationHistory_StartDate.Value},
                {"toDate", activationHistory_EndDate.Value}
            };

            AsyncArgument arg = new AsyncArgument(api, "GetDeviceContractAutoRequests", parameters, typeof(object));
            BeginWorker(arg);
        }

#endregion

#region Device Databases

        private void deviceDatabases_GetOwnDatabasesBtn_Click(object sender, EventArgs e)
        {
            var parameters = new Dictionary<string, object>
            {
                {"apiKey", apiKey},
                {"sessionId", sessionId},
                {"forAccount", deviceDatabases_ForAccountBox.Text}
            };
            AsyncArgument arg = new AsyncArgument(api, "GetOwnDatabases", parameters, typeof(object));
            BeginWorker(arg);
        }

        private void deviceDatabases_GetCurrentBtn_Click(object sender, EventArgs e)
        {
            var parameters = new Dictionary<string, object>
            {
                {"apiKey", apiKey},
                {"sessionId", sessionId},
                {"forAccount", deviceDatabases_ForAccountBox.Text}
            };
            if (!string.IsNullOrEmpty(deviceDatabases_NextId.Text))
            {
                parameters.Add("nextId", Convert.ToInt32(deviceDatabases_NextId.Text));
            }
            AsyncArgument arg = new AsyncArgument(api, "GetCurrentDeviceDatabases", parameters, typeof(object));
            BeginWorker(arg);
        }

#endregion

#region Device Order Entry
        private void deviceOrderEntry_SearchButton_Click(object sender, EventArgs e)
        {
            var forAccount = deviceOrderEntry_ForAccountBox.Text;
            var parameters = new Dictionary<string, object> { { "apiKey", apiKey }, { "sessionId", sessionId }, { "forAccount", forAccount }, { "poNum", deviceOrderEntry_PoNumTextBox.Text.Trim() }, { "includeCarrierInfo", deviceOrderEntry_IncludeCarrierInfoCheckbox.Checked } };
            var asyncArgument = new AsyncArgument(api, "GetDeviceOrderEntries", parameters, typeof(ApiDeviceOrderEntry[]));
            BeginWorker(asyncArgument);
        }

#endregion

#region Helper Methods

        void bw_DoWork(object sender, DoWorkEventArgs e)
        {
            try
            {
                var asyncArgument = (AsyncArgument)e.Argument;

                // to get the native object returned, use this
                //var obj = asyncArgument.Api.Invoke(asyncArgument.MethodName, asyncArgument.ReturnType, asyncArgument.Parameters);

                try
                {
                    // generic object returned, for generic population
                    var obj = asyncArgument.Api.Invoke(asyncArgument.MethodName, typeof(object), asyncArgument.Parameters);
                    e.Result = new AsyncReturnResult(obj, asyncArgument.MethodName);
                }
                catch (SessionExpiredException ex)
                {
                    e.Result = ex;
                }
            }
            catch (Exception ex)
            {
                e.Result = ex;
            }
        }

        void bw_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            LoadResultsToTreeView(treeView, e.Result);
        }

        void LoadResultsToTreeView(TreeView tv, object result)
        {
            tv.Nodes.Clear();
            try
            {
                if (result == null)
                {
                    return;
                }
                // Exception?
                var ex = result as Exception;
                if (ex != null)
                {
                    var n = new TreeNode(ex.StackTrace);
                    tv.Nodes.Add(new TreeNode(ex.Message, new[] { n }));
                    tv.ForeColor = Color.Red;
                    return;
                }
                var simpleStatus = result as string;
                if (!string.IsNullOrEmpty(simpleStatus))
                {
                    tv.Nodes.Add(new TreeNode(simpleStatus));
                    return;
                }
                var returnResult = result as AsyncReturnResult;
                if (returnResult == null)
                {
                    return;
                }
                IList<Dictionary<string, object>> dictionaryItems = null;
                var listOfItems = returnResult.ReturnResult as List<object>;
                if (listOfItems != null && listOfItems.Count > 0)
                {
                    // set columns
                    var firstItem = listOfItems[0] as Dictionary<string, object>;
                    if (firstItem == null)
                    {
                        // single object
                        LoadSingleObjects(tv, listOfItems);
                        return;
                    }
                    foreach (object obj in listOfItems)
                    {
                        var itm = obj as Dictionary<string, object>;
                        if (itm != null)
                        {
                            if (dictionaryItems == null)
                            {
                                dictionaryItems = new List<Dictionary<string, object>>();
                            }
                            dictionaryItems.Add(itm);
                        }
                    }
                }
                else
                {
                    var dis = returnResult.ReturnResult as Dictionary<string, object>;
                    if (dis == null)
                    {
                        LoadSingleObjects(tv, new[] { returnResult });
                        return;
                    }
                    dictionaryItems = new[] { dis };
                }
                LoadDictionaryObjectsToNode(treeView, dictionaryItems, returnResult.ObjectName);
            }
            finally
            {
                treeView.Enabled = true;
                Cursor = Cursors.Default;
            }
        }

        void LoadSingleObjects(TreeView tv, IEnumerable<object> objects)
        {
            foreach (object obj in objects)
            {
                var val = obj.ToString();
                if (string.IsNullOrEmpty(val))
                {
                    continue;
                }
                tv.Nodes.Add(val);
            }
        }

        void LoadDictionaryObjectsToNode(TreeView tv, IList<Dictionary<string, object>> objects, string objectName)
        {
            tv.Nodes.Clear();
            if (objects == null || objects.Count == 0)
            {
                return;
            }
            var counter = 0;
            foreach (Dictionary<string, object> obj in objects)
            {
                var node = SerializeObjectToNode(obj, null);
                node.Text = string.Format("{0} {1}", objectName, counter);
                treeView.Nodes.Add(node);
                counter++;
            }
        }

        TreeNode SerializeObjectToNode(Dictionary<string, object> items, TreeNode previousTreeNode)
        {
            var treeItems = new List<TreeNode>();
            foreach (KeyValuePair<string, object> entry in items)
            {
                var cleanKey = GetCleanKey(entry.Key);
                var dictItem = entry.Value as Dictionary<string, object>;
                if (dictItem == null)
                {
                    // not a json object
                    var listItems = entry.Value as List<object>;
                    if (listItems == null)
                    {
                        var name = string.Format("{0}: {1}", cleanKey, entry.Value);
                        var parent = new TreeNode(name) { Tag = entry.Value };
                        treeItems.Add(parent);
                    }
                    else
                    {
                        foreach (object item in listItems)
                        {
                            var dictItemsFromList = item as Dictionary<string, object>;
                            if (dictItemsFromList != null)
                            {
                                var newNode = new TreeNode(cleanKey);
                                SerializeObjectToNode(dictItemsFromList, newNode);
                                treeItems.Add(newNode);
                            }
                        }
                    }
                }
                else
                {
                    var newNode = new TreeNode(cleanKey);
                    SerializeObjectToNode(dictItem, newNode);
                    treeItems.Add(newNode);
                }
            }
            if (previousTreeNode != null)
            {
                previousTreeNode.Nodes.AddRange(treeItems.ToArray());
                return previousTreeNode;
            }
            return new TreeNode("Item", treeItems.ToArray());
        }

        static string GetCleanKey(string key)
        {
            // set to title case
            key = char.ToUpper(key[0]) + key.Substring(1);
            // Space out Title
            var output = "";
            foreach (var letter in key)
            {
                if (Char.IsUpper(letter) && output.Length > 0)
                {
                    output += " " + letter;
                }
                else
                {
                    output += letter;
                }
            }
            return output;
        }

        void LoadBusyItemToTreeView(TreeView tv)
        {
            tv.Nodes.Clear();
            Cursor = Cursors.AppStarting;
            tv.Nodes.Add("Loading, please wait...");
            tv.Enabled = false;
        }

        void BeginWorker(AsyncArgument asyncArgument)
        {
            LoadBusyItemToTreeView(treeView);
            bw = new BackgroundWorker();
            bw.DoWork += bw_DoWork;
            bw.RunWorkerCompleted += bw_RunWorkerCompleted;
            bw.RunWorkerAsync(asyncArgument);
        }

        private void accountsList_SelectedIndexChanged(object sender, EventArgs e)
        {
            foreach (TabPage tabPage in apiTab.TabPages)
            {
                foreach (Control control in tabPage.Controls)
                {
                    var textBox = control as TextBox;
                    if (textBox == null || textBox.Tag == null || (string)textBox.Tag != "ForAccountBox")
                    {
                        continue;
                    }
                    textBox.Text = accountsList.SelectedItem.ToString();
                }
            }
            try
            {
                LoadAccountSpecificLists();
            }
            catch (SessionExpiredException ex)
            {
                // Session expired, re-authenticate and try again
                if (Authenticate())
                {
                    LoadAccountSpecificLists();
                }
            }

        }


        #endregion

        private void getLastLocation_Click(object sender, EventArgs e)
        {
            int[] hardwareIds = null;

            if (txtHardwareId.Text.Length > 9)
            {
                hardwareIds = Array.ConvertAll(txtHardwareId.Text.Split(','), s=> int.Parse(s));
            }
            else
            {
                int hardwareId = Convert.ToInt32(txtHardwareId.Text.Trim());
                hardwareIds = new int[1];
                hardwareIds[0] = hardwareId;
            }
            
            
            var parameters = new Dictionary<string, object> { { "apiKey", apiKey }, { "sessionId", sessionId }, { "hardwareIds", hardwareIds } };
            var asyncArgument = new AsyncArgument(api, "GetLastDeviceLocation", parameters, typeof(ApiDeviceOrderEntry[]));
            BeginWorker(asyncArgument);
        }
    }
}
